############################################################################
# Matt Rodriguez, LBNL
#Copyright (c) 2003, The Regents of the University of California,
#through Lawrence Berkeley National Laboratory 
#(subject to receipt of any required approvals from the U.S. Dept. of Energy). 
#All rights reserved. 
############################################################################
"""
The interface to the Globus xio module
"""

import xioc
import utilc
from pyGlobus.util import GlobusException

class XioException(GlobusException):
    pass

class XioAttrException(XioException):
    pass

class XioTargetException(XioException):
    pass

class XioStackException(XioException):
    pass

class XioDriverException(XioException):
    pass

class XioServerException(XioException):
    pass

class XioFileException(XioException):
    pass

class XioDataDescriptorException(XioException):
    pass

class XioSocketException(XioException):
    pass

class XioBase:
    """
    Provides functionality all of the xio classes need
    """
    def __init__(self):
        self._handle = None
        ret = utilc.globus_module_activate(xioc.get_module())
        if not ret == 0:
            ex = XioException("Unable to initialize the xio module")
            raise ex
        
    def __del__(self):
        ret = utilc.globus_module_deactivate(xioc.get_module())
        if not ret == 0:
            ex = XioException("Unable to deactivate the xio module")
            raise ex
        
    def get_handle(self):
        return self._handle
    
class XioAttr(XioBase):

    def __init__(self, handle=None):
        """
        Creates an Xio Attr
        """
        XioBase.__init__(self)
        if handle == None:
            ret, handle = xioc.attr_init()
            if not ret == 0:
                raise XioAttrException(handle)

        self._handle = handle

    def __del__(self):
        """
        Frees the memory associated with the XioAttr
        """
        ret, err = xioc.attr_destroy(self._handle)
        if not ret == 0:
            ex = XioAttrException(err)
            raise ex
        XioBase.__del__(self)
        
    def cntl(self, driver, cmd, arg=0):
        """
        Manipulate the values associated with the attribute
        """
        ret, err = xioc.attr_cntl(self._handle, driver.get_handle(),cmd,arg)
        if not ret ==0:
            ex = XioAttrException(err)
	    raise ex
        
    def copy(self):
        """
        Return a new XioAttr object
        """
	ret, newAttr = xioc.attr_copy(self._handle)
        if not ret == 0:
            ex = XioAttrException(newAttr)
            raise ex
        
        return XioAttr(newAttr)   
	

        
class XioDriver(XioBase):
    """
    Creates an XioDriver
    """
    def __init__(self):
        """
        Creates an XioDriver instance
        """
        XioBase.__init__(self)
        self._handle = None
        self._loaded = 0
        
    def __del__(self):
        """
        Frees the memory associated with the XioDriver
        """
        if self._loaded:
            self.unload()
        xioc.driver_free(self._handle)
        XioBase.__del__(self)
        
    def load(self, driverString):
        """
        **Arguments**

        - *driverString* - Valid strings are tcp, gsi
        Should only be called once per instance
        """
        ret, handle = xioc.driver_load(driverString)
        if not ret == 0:
            ex = XioDriverException(handle)
            raise ex
        self._handle = handle
        self._loaded = 1

    def unload(self):
        """
        Unloads the driver, should be called before __del__
        """
        ret, err = xioc.driver_unload(self._handle)
        if not ret == 0:
            ex = XioDriverException(err)
            raise ex
        self._loaded = 0
        
class XioDataDescriptor(XioBase):
    """
    Wraps the globus_xio_data_descriptor_t structure
    """
    def __init__(self, file, handle=None):
        XioBase.__init__(self)
        if handle == None:
            ret, handle = xioc.descriptor_init(file.get_handle())
            if not ret == 0:
                ex = XioDataDescriptorException(handle)
                raise ex
        self._handle = handle

    def __del__(self):
        ret, err = xioc.descriptor_destroy(self._handle)
        if not ret ==0:
            ex = XioDataDescriptorException(err)
            raise ex
        XioBase.__del__(self)
        
class XioStack(XioBase):
    """
    Creates an Xio Attr
    """
    def __init__(self, attr=None, handle=None):
        """
        Creates an XioStack instance
        """
        
        XioBase.__init__(self)
        if attr == None:
            attr = XioAttr()
            
        if handle == None:
            ret, handle = xioc.stack_init(attr.get_handle())
            if not ret == 0:
                raise XioStackException(handle)
            
        self._handle = handle

    def __del__(self):
        """
        Frees the memory associated with the XioTarget
        """   
        ret, err = xioc.stack_destroy(self._handle)
        if not ret == 0:
            ex = XioStackException(err)
            raise ex
        XioBase.__del__(self)
        
    def push_driver(self, driver):
        """
        Pushes a driver onto the Protocol Stack.
        The transport driver must be pushed first

        **Arguments**

        - *driver* - a XioDriver object

        **Returns**
        None

        **Raises**
        A XioStackException is raised if the driver cannot be
        pushed onto the stack
        """
        ret, err = xioc.push_driver(self._handle, driver.get_handle())
        if not ret == 0:
            ex = XioStackException(err)
            raise ex
        
class XioFileBase(XioBase):
    """
    Parent class for XioFile* classes
    """
    def __init__(self,stack= None, handle=None):
        XioBase.__init__(self)
        if handle == None:
            ret, handle = xioc.xio_handle_create(stack.get_handle())
            if not ret == 0:
                ex = XioFileException(handle)
                raise ex
            
        self._handle = handle

    def __del__(self):
        """
        frees the memory malloc'd in either xioc.handle_create
        or in a accept callback,
        """
        xioc.handle_free(self._handle)
        XioBase.__del__(self)
        

    def free_callback1(self, cbHandle):
        xioc.free_callback(cbHandle)

    def free_callback4(self, cbHandle):
        xioc.free_callback4(cbHandle)
        
class XioFile(XioFileBase):
    """
    XioFile class with blocking operations
    """
    
    def __init__(self,stack=None,handle=None):
        XioFileBase.__init__(self,stack,handle)

    def __del__(self):
        XioFileBase.__del__(self)
        
    def open(self, contact_string, attr = None):
        """
        Creates the underlying handle, blocking

        **Arguments**
        - *attr* - a XioAttr object

        - *contact_string* - a Python string

        **Returns**
        None

        **Raises**
        a XioFileException is raised if there is a problem opening the file
        """
        if attr == None:
            attr = XioAttr()
    
        
        ret, err= xioc.xio_open(attr.get_handle(), contact_string, self._handle)
        if not ret ==0:
            ex = XioFileException(err)
            raise ex
        
    def close(self, attr= None):
        """
        **Arguments**
        - *attr* - a XioAttr object

        **Returns**
        None

         **Raises**
        a XioFileException is raised if there is a problem opening the file
        """
        if attr == None:
            attr = XioAttr()
        
        
        ret, err = xioc.xio_close(self._handle, attr.get_handle())
        if not ret ==0:
            ex = XioFileException(err)
            raise ex

    def read(self,buffer, waitForBytes=1,data=None):
        """
        blocking read call

        **Arguments**
        - *buffer* - A pyGlobus.util.Buffer object

        - *waitForBytes* - an int telling how many bytes to wait to recieve before returning

        - *data* - a XioDataDescriptor object

        **Returns**
        A string read from the handle

        **Raises**
        a XioFileException is raised if there is a problem reading from the handle
        """
        if data == None:
            data = XioDataDescriptor(self)
            
        
        ret, outString = xioc.xio_read(self._handle,
                                       buffer.get_handle(),
                                       buffer.get_size(),
                                       waitForBytes,
                                       data.get_handle())

        if not ret == 0:
            ex = XioFileException(outString)
            raise ex

        return outString

    def write(self, writeString, waitForBytes=0, data=None):
        """
        blocking write call

        **Arguments**
        - *writeString* - a string that will be written to the handle

        - *waitForBytes* - an int telling how many bytes to write before returning,
        the default is the length of the string.

        - *data* - a XioDataDescriptor object

        **Returns**
        None

        **Raises**
        a XioFileException is raised if there is a problem writing to the handle
        """
        if data == None:
            data = XioDataDescriptor(self)
            
        if waitForBytes == 0:
            waitForBytes = len(writeString)

        ret, err = xioc.xio_write(self._handle,
                                  writeString,
                                  len(writeString),
                                  waitForBytes,
                                  data.get_handle())
        if not ret == 0:
            ex = XioFileException(err)
            raise ex
        
class XioFileAsyn(XioFileBase):
    """
    XioFile class with Nonblocking operations
    """

    def __init__(self,stack=None, handle=None):
        XioFileBase.__init__(self,stack,handle)
        self._cb1 = []
        self._cb4 = []

    def __del__(self):
        for cb in self._cb1:
            self.free_callback1(cb)
        for cb in self._cb4:
            self.free_callback4(cb)
            
        XioFileBase.__del__(self)
         
    def open(self,callback, arg,contact_string, attr = None):
        """
        Creates the underlying handle, Nonblocking

        **Arguments**
        
        - *callback* - A Python function that has a signature func(arg,handle,errTuple)
        Where arg is the user arg,passed into this function, handle is a SWIG'ized
        globus_xio_handle_t pointer, and errTuple is a tuple containing (0,None) if there
        are no errors or (error_code, error_message) if there is an error 

        - *arg* - A user argument that will be passed to the callback function.

        - *contact_string* - a Python string

        - *attr* - a XioAttr object

        **Returns**
        None

        **Raises**
        a XioFileException is raised if there is a problem opening the file

        """
        if attr == None:
            attr = XioAttr()
    
        
        ret, cbHandle = xioc.xio_register_open(attr.get_handle(),
                                               contact_string,
                                               self._handle,
                                               callback,arg)

        self._cb1.append(cbHandle)
        if not ret ==0:
            ex = XioFileException(cbHandle)
            raise ex
        
        
        
    def close(self, callback, arg, attr = None):
        """
        **Arguments**

        - *attr* - a XioAttr object
        
        - *callback* - A Python function that has a signature func(arg,handle,errTuple)
        Where arg is the user arg,passed into this function, handle is a SWIG'ized
        globus_xio_handle_t pointer, and errTuple is a tuple containing (0,None) if there
        are no errors or (error_code, error_message) if there is an error 

        - *arg* - A user argument that will be passed to the callback function.

        **Returns**
        None

         **Raises**
        a XioFileException is raised if there is a problem opening the file
        """

        if attr ==None:
            attr = XioAttr()
        
        
        ret, cbHandle = xioc.xio_register_close(self._handle,
                                                attr.get_handle(),
                                                callback, arg)

        self._cb1.append(cbHandle)
        if not ret ==0:
            ex = XioFileException(cbHandle)
            raise ex
        
        
    def read(self,callback, arg, buffer, waitForBytes=1,data=None):
        """
        Nonblocking read call

        **Arguments**

        - *callback* - A Python function that has a signature
        func(arg,handle,errTuple,buffer, nbytes, data_desc) Where arg is the user
        arg,passed into this function, handle is a SWIG'ized globus_xio_handle_t
        pointer, and errTuple is a tuple containing (0,None) if there
        are no errors or (error_code, error_message) if there is an error, buffer
        is the data which has been read in, nbytes is the number of bytes read in,
        data_desc is a SWIG'ized globus_xio_data_descriptor_t pointer. 

        - *arg* - A user argument that will be passed to the callback function.
        
        - *buffer* - A pyGlobus.util.Buffer object

        - *waitForBytes* - an int telling how many bytes to wait to recieve before returning

        - *data* - a XioDataDescriptor object

        **Returns**
        None

        **Raises**
        a XioFileException is raised if there is a problem reading from the handle
        """
        
        if data == None:
            data = XioDataDescriptor(self)
            

        ret, cbHandle = xioc.xio_register_read(self._handle,
                                                buffer,
                                                buffer.get_handle(),
                                                buffer.get_size(),
                                                waitForBytes,
                                                data.get_handle(),
                                               callback, arg)

        self._cb4.append(cbHandle)
        if not ret == 0:
            ex = XioFileException(cbHandle)
            raise ex
        

    def write(self, callback, arg,writeString, waitForBytes=0, data=None):
        """
        Nonblocking write call

        **Arguments**
        
        - *callback* - A Python function that has a signature
        func(arg,handle,errTuple,buffer, nbytes, data_desc) Where arg is the user
        arg,passed into this function, handle is a SWIG'ized globus_xio_handle_t
        pointer, and errTuple is a tuple containing (0,None) if there
        are no errors or (error_code, error_message) if there is an error, buffer
        is the data which has been read in, nbytes is the number of bytes read in,
        data_desc is a SWIG'ized globus_xio_data_descriptor_t pointer. 

        - *arg* - A user argument that will be passed to the callback function.
        
        - *writeString* - a string that will be written to the handle

        - *waitForBytes* - an int telling how many bytes to write before returning,
        the default is the length of the string.

        - *data* - a XioDataDescriptor object

        **Returns**
        None

        **Raises**
        a XioFileException is raised if there is a problem writing to the handle
        """
        if data == None:
            data = XioDataDescriptor(self)
        
        if waitForBytes == 0:
            waitForBytes = len(writeString)

        ret, cbHandle = xioc.xio_register_write(self._handle,
                                  writeString,
                                  len(writeString),
                                  waitForBytes,
                                  data.get_handle(),
                                  callback,arg)

        self._cb4.append(cbHandle)
        if not ret == 0:
            ex = XioFileException(cbHandle)
            raise ex

        
class XioServerBase(XioBase):

    def __init__(self,attr, stack, handle=None):
        XioBase.__init__(self)
        if handle == None:
            ret, handle = xioc.server_create(attr.get_handle(),stack.get_handle())
            if not ret == 0:
                ex = XioServerException(handle)
                raise ex
        self._handle = handle
       
    def __del__(self):
        xioc.server_free(self._handle)
        XioBase.__del__(self)

 
    def free_callback1(self, cbHandle):
        xioc.free_callback(cbHandle)
        
    def cntl(self, driver, cmd, arg):
        """
        **Arguments**

        - *driver* - an XioDriver object

        - *cmd* - a command that the driver understands

        - *arg* - an argument for the command that the driver understands, arg must be
        an int for the time being
        """

        ret, err = xioc.server_cntl(self._handle, driver.get_handle(), cmd, arg)
        if not ret == 0:
            ex = XioServerException(err)
     	    raise ex
    
    def cancel_accept(self):
        ret, err = xioc.server_cancel_accept(self._handle)
        if not ret == 0:
            ex = XioServerException(err)
            raise ex

    def get_contact(self):
        ret, contactString = xioc.server_get_contact(self._handle)
        if not ret == 0:
            ex = XioServerException(contactString)
            raise ex
        return contactString
        
class XioServer(XioServerBase):

    def __init__(self, attr, stack, handle = None):
        XioServerBase.__init__(self, attr,stack, handle)
        self._close = 1

    def __del__(self):
        if self._close:
            self.close()
        XioServerBase.__del__(self)
        
    def close(self):
        """
        Close the server, blocking
        """
        ret, err = xioc.server_close(self._handle)
        if not ret == 0:
            ex = XioServerException(err)
            raise ex
        self._close = 0

    def accept(self):
        """
        Accept incoming connections, blocking

        **Arguments**

        **Returns**
        

        **Raises**
        An XioServerException is raised if there is a problem accepting
        a connection
        """
        ret, fileHandle = xioc.server_accept(self._handle)
        if not ret == 0:
            ex = XioServerException(fileHandle)
     	    raise ex
        return XioFile(stack=None,handle=fileHandle)

class XioServerAsyn(XioServerBase):
    """
    XioServer with Nonblocking operations
    """
    def __init__(self, attr, stack, handle = None):
        self._cb1 = []
        XioServerBase.__init__(self,attr, stack, handle)
        

    def __del__(self):
        for cb in self._cb1:
            self.free_callback1(cb)
            
        XioServerBase.__del__(self)

    def accept(self, callback, arg):
        """
        Accept incoming connections, Nonblocking
        
        **Arguments**
        
        - *callback* - a Python function whose signature is
        func(arg, server, handle, errTuple): where arg is the user arg,
        server is a SWIG'ized globus_xio_server_t pointer, target is a
        SWIG'ized globus_xio_handle_t pointer, and errTuple is a tuple
        containing (0,None) if there are no errors or
        (error_code, error_message) if there is an error
        
        - *arg* - a user argument that will be passed to the callback function
        
        **Returns**
        None
        
        **Raises**
        An XioServerException is raised if there is a problem accepting
        a connection
        """
                     
        ret, cbHandle = xioc.server_register_accept(self._handle,
                                                    callback, arg)

        self._cb1.append(cbHandle)
        if not ret == 0:
            ex = XioServerException(cbHandle)
            raise ex 
         
        
    def close(self,callback, arg):
        """
        Close the server, blocking

        **Arguments**

        - *callback* - a Python function whose signature is
        func(arg,server) where arg is the arg you pass into this
        method and server is a SWIG'ized globus_xio_server_t pointer

        - *arg* - a user argument you wish pass to the callback

        **Returns**
        None

        **Raises**
        A XioServerException is raised if there is a problem closing the
        server
        """
        
        ret, cbHandle = xioc.server_register_close(self._handle,callback, arg)
        self._cb1.append(cbHandle)
        if not ret == 0:
            ex = XioServerException(cbHandle)
            raise ex

class XioSocket:
    """
    A simple socket interface similar to the Python Socket API. 
    
    Provides the ability to make strongly authenticated connetions using GSI. It
    supports several modes of authorization and provides optional integrity/privacy
    to the data.   
    """

    def __init__(self, handle = None):
        """
        Constructs an instance.

        Each instance wraps a globus_xio_handle_t object. An instance of this class
        may only have a single open connection at a time. To have more then one open
        connection, use multiple instances.
        
        **Arguments:**
        
        - *handle* An optional argument that should contain a SWIG'ized pointer to
        a globus_xio_handle_t object.
            
        **Raises**        
        Raises a XioSocketException if unable to initialize the globus modules.
        """
        self._driver = XioDriver()
        self._stack = XioStack()
        self._driver.load("tcp")
        self._stack.push_driver(self._driver)
        self._handle = handle
        self._server_handle = None
        self.refcount = 1
        self._attr = XioAttr()
        if handle == None:
            ret, handle = xioc.xio_handle_create(self._stack.get_handle())
            if not ret == 0:
                ex = XioSocketException(handle)
                raise ex
            
        self._handle = handle
        return
    
    def __del__(self):
        pass
    
    def connect(self, host, port):
        """
        Blocking TCP connection establishment.

        Connect a TCP socket on the specified host/port pair.

        **Arguments:**
        
        - *host* A string containing the host to connect to.
        
        - *port* An int containing the port to connect to.
        
        **Raises:**
        
        A XioSocketException is thrown if the handle is currently in use,
        or an error occurs while making the connection.
        """

        contact_string = host +':' + str(port)
        ret, err= xioc.xio_open(self._attr.get_handle(), contact_string, self._handle)
        if not ret ==0:
            ex = XioSocketException(err)
            raise ex

    def create_listener(self, port=0, backlog=-1):
        """
        Creates a listening TCP socket.
        
        Creates a socket handle capable of accepting new TCP connections
        from other hosts or processes. In order to actually listen for
        connections, you must call either listen or register listen on this
        instance.

        **Arguments**
        
        - *port* The TCP port that the socket will listen for connections on. If the
        port number is 0, then an arbitrary TCP port will be selected. If
        this is true, and the restrict_port attribute is set to TRUE (the
        default) and the GLOBUS_TCP_PORT_RANGE environment variable
        was set when Globus XIO was initialized, then the port will be
        selected from that range. Otherwise, any port number may be
        chosen.
        
        - *backlog* The backlog parameter indicates he maximum length of the
        system's queue of pending connections. Any connection attempts
        when the queue is full will fail. If backlog is equal to -1, then the
        system-specific maximum queue length will be used. 
        
        **Raises:**
        
        A XioSocketException is thrown if the handle is currently in use,
        or an error occurs while making the connection.
        
        **Returns:**

        An int containing the port the server will listen on.
        """
            
        if not port == 0:
            self._attr.cntl(self._driver, xioc.GLOBUS_XIO_TCP_SET_PORT, port)

        if not backlog == -1:
            self._attr.cntl(self._driver, xioc.GLOBUS_XIO_TCP_SET_BACKLOG, backlog)
        
        ret, handle = xioc.server_create(self._attr.get_handle(),self._stack.get_handle())
        if not ret == 0:
            ex = XioSocketException(handle)
            raise ex
        
        ret, contactString = xioc.server_get_contact(handle) 
       
        
        self._server_handle = handle

        return contactString.split(':').pop()
    
    def accept(self):
        """
        Blocking server-side TCP connection establishment.
            
        **Raises:**
        
        A XioSocketException is thrown if a listener was not already created, or
        an error occures while doing the underlying accept.
        
        **Returns:**
        
        A new XioSocket object for the new connection.
        """

        ret, childHandle = xioc.server_accept(self._server_handle)
        if not ret == 0:
            ex = XioSocketException(childHandle)
            raise ex
        soc = XioSocket(childHandle)
        return soc
        
    def listen(self):
        """
        Block until a client connection is pending.
        
        This method will block until a connection is pending. Once this method
        returns, accept may be called to return a new GSITCPSocket object.  
        """
        pass
        
    def getpeername(self):
        pass
    
    def getsockname(self):
        pass
    
    def sendall(self, data):
        pass
    
    def send(self, writeString):
        """
        blocking write call

        **Arguments**
        - *writeString* - a string that will be written to the handle 

        **Returns**
        The number of bytes written

        **Raises**
        a XioFileException is raised if there is a problem writing to the handle
        """
    
        dataDesc = XioDataDescriptor(self)

        ret, bytesWritten = xioc.xio_write(self._handle,
                                  writeString,
                                  len(writeString),
                                  1,
                                  dataDesc.get_handle())
        if not ret == 0:
            ex = XioSocketException(bytesWritten)
            raise ex

        return bytesWritten

    
    def read(self,buffer, waitForBytes=1,data=None):
        """
        blocking read call

        **Arguments**
        - *buffer* - A pyGlobus.util.Buffer object

        - *waitForBytes* - an int telling how many bytes to wait to recieve before returning

        - *data* - a XioDataDescriptor object

        **Returns**
        A string read from the handle

        **Raises**
        a XioSocketException is raised if there is a problem reading from the handle
        """
        if data == None:
            data = XioDataDescriptor(self)
            
        
        ret, outString = xioc.xio_read(self._handle,
                                       buffer.get_handle(),
                                       buffer.get_size(),
                                       waitForBytes,
                                       data.get_handle())

        if not ret == 0:
            ex = XioSocketException(outString)
            raise ex

        return outString

    
    def close(self):
        """
        Close this socket, if there are no open XIOFile objects using it.
        """
        if self._server_handle:
            ret, err = xioc.server_close(self._server_handle)
            if not ret == 0:
                ex = XioSocketException(err)
                raise ex
        if self._handle:
            attr = XioAttr()     
            ret, err = xioc.xio_close(self._handle, attr.get_handle())
            if not ret ==0:
                ex = XioSocketException(err)
                raise ex
        
    def settimeout(self, timeout):
        """
        Sets a timeout for the socket 
        """
        ret, err = xioc.attr_cntl_set_timeout_accept(self._attr.get_handle(),
                                                     timeout)
        if not ret == 0:
            ex = XioSocketException(err)
            raise ex
        
        ret, err = xioc.attr_cntl_set_timeout_all(self._attr.get_handle(),
                                                  timeout)
        if not ret == 0:
            ex = XioSocketException(err)
            raise ex
        
    def get_handle(self):
        return self._handle
